from __future__ import print_function

import random
import codecs

import click

from ethereum.tester import (
    accounts,
)

from .server import get_application
from .compat import (
    make_server,
    spawn,
    sleep,
)


@click.command()
@click.option(
    '--host',
    '-h',
    default='localhost',
)
@click.option(
    '--port',
    '-p',
    default=8545,
    type=int,
)
def runserver(host, port):
    application = get_application()

    print(application.rpc_methods.web3_clientVersion())

    print("\nAvailable Accounts\n==================")
    for account in accounts:
        print('0x' + codecs.decode(codecs.encode(account, 'hex'), 'utf8'))

    print("\nListening on %s:%s" % (host, port))

    server = make_server(
        host,
        port,
        application,
    )

    spawn(server.serve_forever)

    try:
        while True:
            sleep(random.random())
    except KeyboardInterrupt:
        try:
            server.stop()
        except AttributeError:
            server.shutdown()


if __name__ == "__main__":
    main()  # noqa



import json
import functools

from werkzeug.wrappers import Request, Response

from jsonrpc import JSONRPCResponseManager, dispatcher

from .rpc import RPCMethods
from .compat import threading
from .client.utils import (
    force_obj_to_text,
)


RESPONSE_HEADERS = {
    "Access-Control-Allow-Headers": "Origin, X-Requested-With, Content-Type, Accept",
    "Access-Control-Allow-Origin": "*",
}


def get_application():
    # When calling server.register_function, first wrap the function in a mutex.
    # This has the effect of serializing all RPC calls. Although we use a
    # multithreaded HTTP server, the EVM itself is not thread-safe, so we must take
    # care when interacting with it.
    rpc_methods = RPCMethods()
    evm_lock = threading.Lock()

    def with_lock(rpc_fn):
        @functools.wraps(rpc_fn)
        def inner(*args, **kwargs):
            evm_lock.acquire()
            try:
                return rpc_fn(*args, **kwargs)
            finally:
                evm_lock.release()

        return inner

    def add_method_with_lock(rpc_fn, *args, **kwargs):
        rpc_fn_with_lock = with_lock(rpc_fn)
        return dispatcher.add_method(rpc_fn_with_lock, *args, **kwargs)

    add_method_with_lock(rpc_methods.eth_coinbase, 'eth_coinbase')
    add_method_with_lock(rpc_methods.eth_accounts, 'eth_accounts')
    add_method_with_lock(rpc_methods.eth_gasPrice, 'eth_gasPrice')
    add_method_with_lock(rpc_methods.eth_blockNumber, 'eth_blockNumber')
    add_method_with_lock(rpc_methods.eth_estimateGas, 'eth_estimateGas')
    add_method_with_lock(rpc_methods.eth_call, 'eth_call')
    add_method_with_lock(rpc_methods.eth_sendTransaction, 'eth_sendTransaction')
    add_method_with_lock(rpc_methods.eth_sendRawTransaction, 'eth_sendRawTransaction')
    add_method_with_lock(rpc_methods.eth_getCompilers, 'eth_getCompilers')
    add_method_with_lock(rpc_methods.eth_compileSolidity, 'eth_compileSolidity')
    add_method_with_lock(rpc_methods.eth_getCode, 'eth_getCode')
    add_method_with_lock(rpc_methods.eth_getBalance, 'eth_getBalance')
    add_method_with_lock(rpc_methods.eth_getTransactionCount, 'eth_getTransactionCount')
    add_method_with_lock(rpc_methods.eth_getTransactionByHash, 'eth_getTransactionByHash')
    add_method_with_lock(rpc_methods.eth_getTransactionReceipt, 'eth_getTransactionReceipt')
    add_method_with_lock(rpc_methods.eth_getBlockByHash, 'eth_getBlockByHash')
    add_method_with_lock(rpc_methods.eth_getBlockByNumber, 'eth_getBlockByNumber')
    add_method_with_lock(rpc_methods.eth_newBlockFilter, 'eth_newBlockFilter')
    add_method_with_lock(rpc_methods.eth_newPendingTransactionFilter,
                         'eth_newPendingTransactionFilter')
    add_method_with_lock(rpc_methods.eth_newFilter, 'eth_newFilter')
    add_method_with_lock(rpc_methods.eth_getFilterChanges, 'eth_getFilterChanges')
    add_method_with_lock(rpc_methods.eth_getFilterLogs, 'eth_getFilterLogs')
    add_method_with_lock(rpc_methods.eth_uninstallFilter, 'eth_uninstallFilter')
    add_method_with_lock(rpc_methods.eth_protocolVersion, 'eth_protocolVersion')
    add_method_with_lock(rpc_methods.eth_syncing, 'eth_syncing')
    add_method_with_lock(rpc_methods.eth_mining, 'eth_mining')
    add_method_with_lock(rpc_methods.web3_sha3, 'web3_sha3')
    add_method_with_lock(rpc_methods.web3_clientVersion, 'web3_clientVersion')
    add_method_with_lock(rpc_methods.net_version, 'net_version')
    add_method_with_lock(rpc_methods.net_listening, 'net_listening')
    add_method_with_lock(rpc_methods.net_peerCount, 'net_peerCount')
    add_method_with_lock(rpc_methods.evm_reset, 'evm_reset')
    add_method_with_lock(rpc_methods.evm_snapshot, 'evm_snapshot')
    add_method_with_lock(rpc_methods.evm_revert, 'evm_revert')
    add_method_with_lock(rpc_methods.evm_mine, 'evm_mine')
    add_method_with_lock(rpc_methods.rpc_configure, 'rpc_configure')
    add_method_with_lock(rpc_methods.testing_timeTravel, 'testing_timeTravel')
    add_method_with_lock(rpc_methods.personal_importRawKey, 'personal_importRawKey')
    add_method_with_lock(rpc_methods.personal_listAccounts, 'personal_listAccounts')
    add_method_with_lock(rpc_methods.personal_lockAccount, 'personal_lockAccount')
    add_method_with_lock(rpc_methods.personal_newAccount, 'personal_newAccount')
    add_method_with_lock(rpc_methods.personal_unlockAccount, 'personal_unlockAccount')
    add_method_with_lock(rpc_methods.personal_signAndSendTransaction,
                         'personal_signAndSendTransaction')

    @Request.application
    def application(request):
        response = JSONRPCResponseManager.handle(
            request.data,
            dispatcher,
        )
        response = Response(
            json.dumps(force_obj_to_text(response.data, True)),
            headers=RESPONSE_HEADERS,
            mimetype='application/json',
        )
        return response

    application.rpc_methods = rpc_methods

    return application



import rlp
from rlp.utils import (
    int_to_big_endian,
)

from ethereum.utils import zpad

from .utils import (
    encode_32bytes,
    encode_number,
    encode_address,
    encode_data,
)


def serialize_txn_receipt(block, txn, txn_index):
    txn_receipt = block.get_receipt(txn_index)
    origin_gas = block.transaction_list[0].startgas

    if txn.creates is not None:
        contract_addr = encode_address(txn.creates)
    else:
        contract_addr = None

    return {
        "transactionHash": encode_32bytes(txn.hash),
        "transactionIndex": encode_number(txn_index),
        "blockNumber": encode_number(block.number),
        "blockHash": encode_32bytes(block.hash),
        "cumulativeGasUsed": encode_number(origin_gas - txn.startgas + txn_receipt.gas_used),
        "gasUsed": encode_number(txn_receipt.gas_used),
        "contractAddress": contract_addr,
        "logs": [
            serialize_log(block, txn, txn_index, log, log_index)
            for log_index, log in enumerate(txn_receipt.logs)
        ],
    }


def serialize_txn(block, txn, txn_index):
    return {
        "hash": encode_32bytes(txn.hash),
        "nonce": encode_number(txn.nonce),
        "blockHash": encode_32bytes(block.hash),
        "blockNumber": encode_number(block.number),
        "transactionIndex": encode_number(txn_index),
        "from": encode_address(txn.sender),
        "to": encode_address(txn.to),
        "value": encode_number(txn.value),
        "gas": encode_number(txn.startgas),
        "gasPrice": encode_number(txn.gasprice),
        "input": encode_data(txn.data)
    }


def serialize_log(block, txn, txn_index, log, log_index):
    return {
        "type": "mined",
        "logIndex": encode_number(log_index),
        "transactionIndex": encode_number(txn_index),
        "transactionHash": encode_32bytes(txn.hash),
        "blockHash": encode_32bytes(block.hash),
        "blockNumber": encode_number(block.number),
        "address": encode_32bytes(log.address),
        "data": encode_data(log.data),
        "topics": [
            encode_number(topic, 32) for topic in log.topics
        ],
    }


def serialize_block(block, full_transactions):
    if full_transactions:
        transactions = [
            serialize_txn(block, txn, txn_index)
            for txn_index, txn in enumerate(block.transaction_list)
        ]
    else:
        transactions = [encode_32bytes(txn.hash) for txn in block.transaction_list]

    unpadded_logs_bloom = int_to_big_endian(block.bloom)
    logs_bloom = zpad(unpadded_logs_bloom, (256 - len(unpadded_logs_bloom)))

    return {
        "number": encode_number(block.number),
        "hash": encode_32bytes(block.hash),
        "parentHash": encode_32bytes(block.prevhash),
        "nonce": encode_32bytes(block.nonce),
        "sha3Uncles": encode_32bytes(block.uncles_hash),
        # TODO logsBloom / padding
        "logsBloom": logs_bloom,
        "transactionsRoot": encode_32bytes(block.tx_list_root),
        "stateRoot": encode_32bytes(block.state_root),
        "miner": encode_address(block.coinbase),
        "difficulty": encode_number(block.difficulty),
        "totalDifficulty": encode_number(block.chain_difficulty()),
        "size": encode_number(len(rlp.encode(block))),
        "extraData": encode_32bytes(block.extra_data),
        "gasLimit": encode_number(block.gas_limit),
        "gasUsed": encode_number(block.gas_used),
        "timestamp": encode_number(block.timestamp),
        "transactions": transactions,
        "uncles": block.uncles
    }



import sys
import logging

import rlp

try:
    from sha3 import keccak_256
except ImportError:
    from sha3 import sha3_256 as keccak_256

from ethereum import blocks
from ethereum.tester import (
    languages,
)

from .client import EthTesterClient
from .client.utils import (
    force_bytes,
    decode_hex,
    encode_number,
    encode_32bytes,
)

from .utils import (
    input_transaction_formatter,
    input_filter_params_formatter,
    normalize_block_number,
)


logger = logging.getLogger(__name__)


class RPCMethods(object):
    def __init__(self):
        self.full_reset()

    def full_reset(self):
        self.client = EthTesterClient()
        self.reset_rpc_meta()

    def reset_rpc_meta(self):
        self.RPC_META = {
            'eth_protocolVersion': 63,
            'eth_syncing': False,
            'eth_mining': False,
            'net_version': 1,
            'net_listening': False,
            'net_peerCount': 0,
            'homestead_block_number': self.client.evm.block.config['HOMESTEAD_FORK_BLKNUM'],
            'dao_fork_block_number': self.client.evm.block.config['DAO_FORK_BLKNUM'],
            'anti_dos_fork_block_number': self.client.evm.block.config['ANTI_DOS_FORK_BLKNUM'],
            'clearing_fork_block_number': self.client.evm.block.config['CLEARING_FORK_BLKNUM'],
        }

    def rpc_configure(self, key, value):
        self.RPC_META[key] = value

        if key == 'homestead_block_number':
            self.client.evm.block.config['HOMESTEAD_FORK_BLKNUM'] = value
        elif key == 'dao_fork_block_number':
            self.client.evm.block.config['DAO_FORK_BLKNUM'] = value
        elif key == 'anti_dos_fork_block_number':
            self.client.evm.block.config['ANTI_DOS_FORK_BLKNUM'] = value
        elif key == 'clearing_fork_block_number':
            self.client.evm.block.config['CLEARING_FORK_BLKNUM'] = value

    #
    # Snapshot and Reset
    #
    def evm_reset(self):
        self.client.reset_evm()
        return True

    def evm_snapshot(self):
        snapshot_idx = self.client.snapshot_evm()
        return encode_number(snapshot_idx)

    def evm_revert(self, snapshot_idx=None):
        try:
            self.client.revert_evm(snapshot_idx)
        except ValueError:
            return False
        else:
            return True

    def evm_mine(self, num_blocks=1):
        for _ in range(num_blocks):
            self.client.mine_block()

    #
    #  Timetravel
    #
    def testing_timeTravel(self, timestamp):
        if timestamp <= self.client.evm.block.timestamp:
            raise ValueError(
                "Cannot travel back in time.  You'll disrupt the space time "
                "continuum"
            )
        self.client.evm.block.finalize()
        self.client.evm.block.commit_state()
        self.client.evm.db.put(
            self.client.evm.block.hash,
            rlp.encode(self.client.evm.block),
        )

        block = blocks.Block.init_from_parent(
            self.client.evm.block,
            decode_hex(self.eth_coinbase()),
            timestamp=timestamp,
        )

        self.client.evm.block = block
        self.client.evm.blocks.append(block)
        return timestamp

    #
    #  eth_ Functions
    #
    def eth_coinbase(self):
        logger.info('eth_coinbase')
        return self.client.get_coinbase()

    def eth_gasPrice(self):
        logger.info('eth_gasPrice')
        return encode_number(self.client.get_gas_price())

    def eth_blockNumber(self):
        logger.info('eth_blockNumber')
        return encode_number(self.client.get_block_number())

    def eth_sendTransaction(self, transaction):
        formatted_transaction = input_transaction_formatter(transaction)
        return self.client.send_transaction(**formatted_transaction)

    def eth_estimateGas(self, transaction, block_number="latest"):
        formatted_transaction = input_transaction_formatter(transaction)
        return self.client.estimate_gas(**formatted_transaction)

    def eth_sendRawTransaction(self, raw_tx):
        return self.client.send_raw_transaction(raw_tx)

    def eth_call(self, transaction, block_number="latest"):
        formatted_transaction = input_transaction_formatter(transaction)
        return self.client.call(**formatted_transaction)

    def eth_accounts(self):
        return self.client.get_accounts()

    def eth_getCompilers(self):
        return languages.keys()

    def eth_compileSolidity(self, code):
        # TODO: convert to python-solidity lib once it exists
        raise NotImplementedError("This has not yet been implemented")

    def eth_getCode(self, address, block_number="latest"):
        return self.client.get_code(address, normalize_block_number(block_number))

    def eth_getBalance(self, address, block_number="latest"):
        return self.client.get_balance(address, normalize_block_number(block_number))

    def eth_getTransactionCount(self, address, block_number="latest"):
        return encode_number(self.client.get_transaction_count(
            address,
            normalize_block_number(block_number),
        ))

    def eth_getTransactionByHash(self, tx_hash):
        try:
            return self.client.get_transaction_by_hash(tx_hash)
        except ValueError:
            return None

    def eth_getBlockByHash(self, block_hash, full_tx=True):
        return self.client.get_block_by_hash(block_hash, full_tx)

    def eth_getBlockByNumber(self, block_number, full_tx=True):
        return self.client.get_block_by_number(normalize_block_number(block_number), full_tx)

    def eth_getTransactionReceipt(self, tx_hash):
        try:
            return self.client.get_transaction_receipt(tx_hash)
        except ValueError:
            return None

    def eth_newBlockFilter(self, *args, **kwargs):
        # TODO: convert to tester_client once filters implemented
        raise NotImplementedError("This has not yet been implemented")

    def eth_newPendingTransactionFilter(self, *args, **kwargs):
        # TODO: convert to tester_client once filters implemented
        raise NotImplementedError("This has not yet been implemented")

    def eth_newFilter(self, filter_params):
        formatted_filter_params = input_filter_params_formatter(filter_params)
        return self.client.new_filter(**formatted_filter_params)

    def eth_getFilterChanges(self, filter_id):
        return self.client.get_filter_changes(filter_id)

    def eth_getFilterLogs(self, filter_id):
        return self.client.get_filter_logs(filter_id)

    def eth_uninstallFilter(self, filter_id):
        return self.client.uninstall_filter(filter_id)

    def eth_protocolVersion(self):
        return self.RPC_META['eth_protocolVersion']

    def eth_syncing(self):
        return self.RPC_META['eth_syncing']

    def eth_mining(self):
        return self.RPC_META['eth_mining']

    def web3_sha3(self, value, encoding='hex'):
        logger.info('web3_sha3')
        if encoding == 'hex':
            value = decode_hex(value)
        else:
            value = force_bytes(value)
        return encode_32bytes(keccak_256(value).digest())

    @classmethod
    def web3_clientVersion(cls):
        from testrpc import __version__
        return "TestRPC/" + __version__ + "/{platform}/python{v.major}.{v.minor}.{v.micro}".format(
            v=sys.version_info,
            platform=sys.platform,
        )

    #
    # net_ API
    #
    def net_version(self):
        return self.RPC_META['net_version']

    def net_listening(self):
        return self.RPC_META['net_listening']

    def net_peerCount(self):
        return self.RPC_META['net_peerCount']

    #
    # personal_ API
    #
    def personal_listAccounts(self):
        return self.eth_accounts()

    def personal_importRawKey(self, private_key, passphrase):
        return self.client.import_raw_key(private_key, passphrase)

    def personal_lockAccount(self, address):
        return self.client.lock_account(address)

    def personal_newAccount(self, passphrase=None):
        return self.client.new_account(passphrase)

    def personal_unlockAccount(self, address, passphrase, duration=None):
        return self.client.unlock_account(address, passphrase, duration)

    def personal_signAndSendTransaction(self, transaction, passphrase):
        formatted_transaction = input_transaction_formatter(transaction)
        return self.client.send_and_sign_transaction(passphrase, **formatted_transaction)



"""
Large portions of this implementation were influenced or downright copied from
the `eth-testrpc` project by ConsenSys

https://github.com/ConsenSys/eth-testrpc
"""
import time
import itertools
import functools

from testrpc.compat import threading

import rlp

from ethereum import transactions
from ethereum import tester as t
from ethereum.utils import (
    privtoaddr,
)

from .utils import (
    is_string,
    coerce_args_to_bytes,
    strip_0x,
    encode_32bytes,
    encode_address,
    encode_data,
    normalize_number,
    normalize_address,
    decode_hex,
    mk_random_privkey,
    is_array,
    normalize_block_identifier)
from .serializers import (
    serialize_txn,
    serialize_txn_receipt,
    serialize_block,
)
from .filters import (
    check_filter_topics_validity,
    process_block,
    get_filter_bounds,
)


# Set the gas
DEFAULT_GAS_LIMIT = t.gas_limit = t.GAS_LIMIT = 3141592


def with_lock(client_method):
    @functools.wraps(client_method)
    def inner(self, *args, **kwargs):
        self._evm_lock.acquire()
        try:
            return client_method(self, *args, **kwargs)
        finally:
            self._evm_lock.release()

    return inner


class EthTesterClient(object):
    """
    Stand-in replacement for the rpc client that speaks directly to the
    `ethereum.tester` facilities.
    """
    locked_accounts = None
    homestead_block_number = 0
    dao_fork_block_number = 0
    anti_dos_fork_block_number = 0
    clearing_fork_block_number = 0
    dao_fork_support = True

    def __init__(self):
        self._evm_lock = threading.Lock()

        self.snapshots = []

        self.reset_evm()

        self.evm.block.config['HOMESTEAD_FORK_BLKNUM'] = self.homestead_block_number  # noqa
        self.evm.block.config['DAO_FORK_BLKNUM'] = self.dao_fork_block_number
        self.evm.block.config['ANTI_DOS_FORK_BLKNUM'] = self.anti_dos_fork_block_number
        self.evm.block.config['CLEARING_FORK_BLKNUM'] = self.clearing_fork_block_number

        self.passphrase_accounts = {}
        self.passphrase_account_keys = {}
        self.unlocked_accounts = {}

        self.log_filters = {}
        self.log_filters_id_generator = itertools.count()

    @with_lock
    def reset_evm(self, snapshot_idx=None):
        if snapshot_idx is not None:
            self.revert_evm(snapshot_idx)
        else:
            self.evm = t.state()
            self.evm.block.gas_limit = DEFAULT_GAS_LIMIT

    @with_lock
    def snapshot_evm(self):
        self.snapshots.append((self.evm.block.number, self.evm.snapshot()))
        return len(self.snapshots) - 1

    @with_lock
    def revert_evm(self, snapshot_idx=None, reset_logs=False):
        if len(self.snapshots) == 0:
            raise ValueError("No snapshots to revert to")

        if snapshot_idx is not None:
            block_number, snapshot = self.snapshots.pop(snapshot_idx)
        else:
            block_number, snapshot = self.snapshots.pop()

        # Remove all blocks after our saved block number.
        del self.evm.blocks[block_number:]

        self.evm.revert(snapshot)
        self.evm.blocks.append(self.evm.block)

    @with_lock
    def mine_block(self):
        self.evm.mine()

    def wait_for_block(self, block_number, max_wait=0):
        while self.get_block_number() < block_number:
            self.mine_block()
        return self.get_block_by_number(self.get_block_number())

    def wait_for_transaction(self, txn_hash, max_wait=0):
        return self.get_transaction_receipt(txn_hash)

    def get_max_gas(self):
        return t.gas_limit

    #
    # Internal getters for EVM objects
    #
    def _get_transaction_by_hash(self, txn_hash):
        txn_hash = strip_0x(txn_hash)
        if len(txn_hash) == 64:
            txn_hash = decode_hex(txn_hash)
        for block in reversed(self.evm.blocks):
            txn_hashes = block.get_transaction_hashes()

            if txn_hash in txn_hashes:
                txn_index = txn_hashes.index(txn_hash)
                txn = block.transaction_list[txn_index]
                break
        else:
            raise ValueError("Transaction not found")
        return block, txn, txn_index

    def _get_block_by_number(self, block_number="latest"):
        if block_number == "latest":
            if not self.evm.blocks:
                raise ValueError("No blocks")
            elif len(self.evm.blocks) > 1:
                return self.evm.blocks[-2]
            else:
                return self.evm.blocks[-1]
        elif block_number == "earliest":
            return self.evm.blocks[0]
        elif block_number == "pending":
            return self.evm.block
        else:
            block_number = normalize_number(block_number)

            if block_number >= len(self.evm.blocks):
                raise ValueError("Invalid block number")
            return self.evm.blocks[block_number]

    def _get_block_by_hash(self, block_hash):
        if len(block_hash) > 32:
            block_hash = decode_hex(strip_0x(block_hash))
        for block in self.evm.blocks:
            if block.hash == block_hash:
                return block
        else:
            raise ValueError("Could not find block for provided hash")

    @with_lock
    @coerce_args_to_bytes
    def _send_transaction(self, _from=None, to=None, gas=None, gas_price=None,
                          value=0, data=b''):
        """
        The tester doesn't care about gas so we discard it.
        """
        if _from is None:
            _from = self.get_coinbase()

        _from = normalize_address(_from)

        try:
            sender = t.keys[t.accounts.index(_from)]
        except ValueError:
            if _from in self.unlocked_accounts:
                unlock_expiration = self.unlocked_accounts[_from]
                if unlock_expiration is None or unlock_expiration > time.time():
                    sender = self.passphrase_account_keys[_from]
                else:
                    raise ValueError("Account locked.  Unlock before sending tx")
            else:
                raise

        if to is None:
            to = b''

        to = normalize_address(to, allow_blank=True)

        if data is None:
            data = b''

        data = decode_hex(data)

        try:
            if gas is not None:
                t.gas_limit = gas
            if gas_price is not None:
                t.gas_price = gas_price
            output = self.evm.send(sender=sender, to=to, value=value, evmdata=data)
        finally:
            t.gas_limit = t.GAS_LIMIT
            t.gas_price = t.GAS_PRICE

        return output

    #
    # Public API
    #
    def get_coinbase(self):
        return encode_address(self.evm.block.coinbase)

    def get_accounts(self):
        return [
            encode_address(addr) for addr in t.accounts
        ] + [
            encode_address(addr) for addr in self.passphrase_accounts.keys()
        ]

    def get_code(self, address, block_number="latest"):
        block = self._get_block_by_number(block_number)
        return encode_data(block.get_code(strip_0x(address)))

    def estimate_gas(self, *args, **kwargs):
        snapshot_idx = self.snapshot_evm()

        try:
            txn_hash = self.send_transaction(*args, **kwargs)
            txn_receipt = self.get_transaction_receipt(txn_hash)
            gas_used = txn_receipt['gasUsed']
        finally:
            self.revert_evm(snapshot_idx)

        return gas_used

    def send_transaction(self, *args, **kwargs):
        try:
            self._send_transaction(*args, **kwargs)
        finally:
            self.mine_block()
        return encode_32bytes(self.evm.last_tx.hash)

    def send_raw_transaction(self, raw_tx):
        tx = rlp.decode(decode_hex(strip_0x(raw_tx)), transactions.Transaction)

        to = encode_address(tx.to) if tx.to else b''
        _from = encode_address(tx.sender)
        data = encode_data(tx.data)

        return self.send_transaction(
            _from=_from,
            to=to,
            gas_price=tx.gasprice,
            gas=tx.startgas,
            value=tx.value,
            data=data,
        )

    def get_transaction_receipt(self, txn_hash):
        block, txn, txn_index = self._get_transaction_by_hash(txn_hash)

        return serialize_txn_receipt(block, txn, txn_index)

    def get_transaction_count(self, address, block_number="latest"):
        block = self._get_block_by_number(block_number)
        address = normalize_address(address)

        return block.get_nonce(address)

    def get_block_by_number(self, block_number, full_transactions=True):
        block = self._get_block_by_number(block_number)

        return serialize_block(block, full_transactions)

    def get_block_by_hash(self, block_hash, full_transactions=True):
        block = self._get_block_by_hash(block_hash)

        return serialize_block(block, full_transactions)

    def get_block_number(self):
        if len(self.evm.blocks) < 2:
            return 0
        else:
            return self.evm.blocks[-2].number

    def get_gas_price(self):
        return t.gas_price

    def get_balance(self, address, block="latest"):
        _block = self._get_block_by_number(block)
        return _block.get_balance(strip_0x(address))

    def call(self, *args, **kwargs):
        if len(args) >= 7 and args[6] != "latest":
            raise ValueError("Using call on any block other than latest is unsupported")
        if kwargs.get('block', 'latest') != "latest":
            raise ValueError("Using call on any block other than latest is unsupported")
        snapshot_idx = self.snapshot_evm()
        output = self._send_transaction(*args, **kwargs)
        self.revert_evm(snapshot_idx)

        return encode_data(output)

    def get_transaction_by_hash(self, txn_hash):
        block, txn, txn_index = self._get_transaction_by_hash(txn_hash)
        return serialize_txn(block, txn, txn_index)

    def lock_account(self, address):
        address = normalize_address(address)
        self.unlocked_accounts.pop(address, None)
        return True

    @coerce_args_to_bytes
    def check_passphrase(self, address, passphrase):
        address = normalize_address(address)
        if address not in self.passphrase_accounts:
            return False
        elif passphrase == self.passphrase_accounts[address]:
            return True
        else:
            return False

    @coerce_args_to_bytes
    def unlock_account(self, address, passphrase, duration=None):
        address = normalize_address(address)
        if self.check_passphrase(address, passphrase):
            if duration is not None:
                unlock_expiration = time.time() + duration
            else:
                unlock_expiration = None
            self.unlocked_accounts[address] = unlock_expiration
            return True
        return False

    @coerce_args_to_bytes
    def import_raw_key(self, private_key, passphrase):
        if not passphrase:
            raise ValueError("Cannot have empty passphrase")

        public_key = privtoaddr(private_key)

        self.passphrase_accounts[public_key] = passphrase
        self.passphrase_account_keys[public_key] = private_key

        return encode_address(public_key)

    @coerce_args_to_bytes
    def new_account(self, passphrase):
        private_key = mk_random_privkey()

        return self.import_raw_key(private_key, passphrase)

    @coerce_args_to_bytes
    def send_and_sign_transaction(self, passphrase, **txn_kwargs):
        try:
            _from = txn_kwargs['_from']
        except KeyError:
            raise KeyError("`send_and_sign_transaction` requires a `_from` address to be specified")  # NOQa

        _from = normalize_address(_from)

        try:
            self.unlock_account(_from, passphrase)
            return self.send_transaction(**txn_kwargs)
        finally:
            self.lock_account(_from)

    log_filters = None

    def new_filter(self, from_block=None, to_block=None, address=None, topics=None):
        if topics is None:
            topics = []

        if address is None:
            address = []

        try:
            from_block = normalize_block_identifier(from_block)
        except ValueError:
            raise ValueError("Invalid filter parameter for `from_block`")

        try:
            to_block = normalize_block_identifier(to_block)
        except ValueError:
            raise ValueError("Invalid filter parameter for `to_block`")

        if is_string(address):
            addresses = [encode_32bytes(normalize_address(address))]
        elif is_array(address):
            if not address:
                addresses = []
            elif all(is_string(addr) for addr in address):
                addresses = [encode_32bytes(normalize_address(addr)) for addr in address]
            else:
                raise ValueError(
                    "Address must be either a single address or an array of addresses"
                )
        else:
            raise ValueError(
                "Address must be either a single address or an array of addresses"
            )

        if not check_filter_topics_validity(topics):
            raise ValueError("Invalid topics")

        log_filter = {
            'from_block': from_block,
            'to_block': to_block,
            'addresses': addresses,
            'topics': topics,
        }
        filter_id = next(self.log_filters_id_generator)
        self.log_filters[filter_id] = log_filter
        return filter_id

    def new_block_filter(self, *args, **kwargs):
        raise NotImplementedError("TODO")

    def new_pending_transaction_filter(self, *args, **kwargs):
        raise NotImplementedError("TODO")

    def uninstall_filter(self, filter_id):
        try:
            self.log_filters.pop(filter_id)
        except KeyError:
            return False
        else:
            return True

    def get_filter_changes(self, filter_id):
        try:
            log_filter = self.log_filters[filter_id]
        except KeyError:
            raise ValueError("Filter not found for id: {0}".format(filter_id))

        bookmark = log_filter.get('bookmark', None)
        block_slice = get_filter_bounds(
            log_filter['from_block'],
            log_filter['to_block'],
            bookmark,
        )

        block_processor_fn = functools.partial(
            process_block,
            from_block=log_filter['from_block'],
            to_block=log_filter['to_block'],
            addresses=log_filter['addresses'],
            filter_topics=log_filter['topics'],
        )

        log_changes = list(itertools.chain.from_iterable((
            block_processor_fn(block) for block in self.evm.blocks[block_slice]
        )))

        self.log_filters[filter_id]['bookmark'] = self.evm.blocks[-1].number
        return log_changes

    def get_filter_logs(self, filter_id):
        try:
            log_filter = self.log_filters[filter_id]
        except KeyError:
            raise ValueError("Filter not found for id: {0}".format(filter_id))

        block_slice = get_filter_bounds(
            log_filter['from_block'],
            log_filter['to_block'],
        )

        block_processor_fn = functools.partial(
            process_block,
            from_block=log_filter['from_block'],
            to_block=log_filter['to_block'],
            addresses=log_filter['addresses'],
            filter_topics=log_filter['topics'],
        )

        all_log_entries = list(itertools.chain.from_iterable((
            block_processor_fn(block) for block in self.evm.blocks[block_slice]
        )))

        return all_log_entries



def normalize_number(value):
    try:
        return int(value, 16)
    except TypeError:
        return int(value)


def normalize_block_number(value):
    try:
        return normalize_number(value)
    except ValueError:
        return value


def noop(value):
    return value


TXN_KWARGS_MAP = {
    'from': '_from',
    'gasPrice': 'gas_price',
}


TXN_FORMATTERS = {
    'value': normalize_number,
    'gas': normalize_number,
    'gasPrice': normalize_number
}


def input_transaction_formatter(transaction):
    return {
        TXN_KWARGS_MAP.get(k, k): TXN_FORMATTERS.get(k, noop)(v)
        for k, v in transaction.items()
    }


FILTER_KWARGS_MAP = {
    'fromBlock': 'from_block',
    'toBlock': 'to_block',
}


FILTER_FORMATTERS = {
    'fromBlock': normalize_block_number,
    'toBlock': normalize_block_number,
}


def input_filter_params_formatter(filter_params):
    return {
        FILTER_KWARGS_MAP.get(k, k): FILTER_FORMATTERS.get(k, noop)(v)
        for k, v in filter_params.items()
    }



import pkg_resources


try:
    from sha3 import keccak_256  # NOQA
except ImportError:
    pass
else:
    import sha3
    sha3.sha3_256 = sha3.keccak_256


__version__ = pkg_resources.get_distribution("eth-testrpc").version



import functools

from .utils import (
    is_string,
    is_numeric,
    is_array,
    coerce_args_to_bytes,
)
from .serializers import (
    serialize_log,
)


def is_empty_array(value):
    return value == [] or value == tuple()


def is_topic_array(value):
    if not is_array(value):
        return False
    return all(is_string(item) or item is None for item in value)


def is_nested_topic_array(value):
    if not is_array(value):
        return False

    return all((is_topic_array(item) for item in value))


def check_filter_topics_validity(filter_topics):
    return any((
        is_empty_array(filter_topics),
        is_topic_array(filter_topics),
        is_nested_topic_array(filter_topics),
    ))


@coerce_args_to_bytes
def check_topic_match(filter_topic, log_topic):
    if filter_topic is None:
        return True
    return filter_topic == log_topic


@coerce_args_to_bytes
def check_if_topics_match(filter_topics, log_topics):
    if is_empty_array(filter_topics):
        return True
    elif is_topic_array(filter_topics):
        if len(filter_topics) > len(log_topics):
            return False
        return all(
            check_topic_match(filter_topic, log_topic)
            for filter_topic, log_topic
            in zip(filter_topics, log_topics)
        )
    elif is_nested_topic_array(filter_topics):
        return any(
            check_if_topics_match(sub_topics, log_topics)
            for sub_topics in filter_topics
        )
    else:
        raise ValueError("Invalid filter topics format")


@coerce_args_to_bytes
def check_if_log_matches(log_entry, from_block, to_block,
                         addresses, filter_topics):
    log_block_number = int(log_entry['blockNumber'], 16)

    #
    # validate `from_block` (left bound)
    #
    if from_block is None or is_string(from_block):
        pass
    elif is_numeric(from_block):
        if from_block > log_block_number:
            return False
    else:
        raise TypeError("Invalid `from_block`")

    #
    # validate `to_block` (left bound)
    #
    if to_block is None or is_string(to_block):
        pass
    elif is_numeric(to_block):
        if to_block < log_block_number:
            return False
    else:
        raise TypeError("Invalid `to_block`")

    if log_entry['type'] == "pending":
        if to_block != "pending":
            return False
    elif from_block == "pending":
        return False

    #
    # validate `addresses`
    #
    if addresses and log_entry['address'] not in addresses:
        return False

    #
    # validate `topics`
    if not check_if_topics_match(filter_topics, log_entry['topics']):
        return False

    return True


def process_block(block, from_block, to_block, addresses, filter_topics):
    is_filter_match_fn = functools.partial(
        check_if_log_matches,
        from_block=from_block,
        to_block=to_block,
        addresses=addresses,
        filter_topics=filter_topics,
    )

    # TODO: this is really inneficient since many of the early exit conditions
    # can be identified prior to serializing the log entry.  Revamp this so
    # that the functionality in `check_if_log_matches` is more granular and
    # each piece can be checked at the earliers entry point.

    for txn_index, txn in enumerate(block.transaction_list):
        txn_receipt = block.get_receipt(txn_index)
        for log_index, log in enumerate(txn_receipt.logs):
            log_entry = serialize_log(block, txn, txn_index, log, log_index)
            if is_filter_match_fn(log_entry):
                yield log_entry


def get_filter_bounds(from_block, to_block, bookmark=None):
    if bookmark is not None:
        left_bound = bookmark
    elif from_block is None:
        left_bound = None
    elif from_block == "latest":
        left_bound = -1
    elif from_block == "earliest":
        left_bound = None
    elif from_block == "pending":
        left_bound = None
    else:
        left_bound = from_block

    if to_block is None:
        right_bound = None
    elif to_block == "latest":
        right_bound = None
    elif to_block == "earliest":
        right_bound = 1
    elif to_block == "pending":
        right_bound = None
    else:
        right_bound = to_block + 1

    return slice(left_bound, right_bound)



"""An Ethereum client simulator that provides instant results and quick
feedback during development involving smart contracts.

https://github.com/pipermerriam/eth-testrpc
"""

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='eth-testrpc',
    version="1.1.0",
    description='An Ethereum simulator for aiding smart contract development.',
    long_description=long_description,
    url='https://github.com/pipermerriam/eth-testrpc',
    author='Piper Merriam',
    author_email='pipermerriam@gmail.com',
    license='MIT',
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='ethereum blockchain development testing',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=[
        'Werkzeug>=0.11.10',
        'click>=6.6',
        'ethereum>=1.5.2',
        'json-rpc>=1.10.3',
        'rlp>=0.4.4,<0.4.7',
    ],
    extras_require={
        'gevent': [
            "gevent>=1.1.1,<1.2.0",
        ],
    },
    entry_points={
        'console_scripts': [
            'testrpc-py=testrpc.cli:runserver',
        ],
    }
)



import sys
import functools
import random
import codecs

from rlp.utils import (
    int_to_big_endian,
    big_endian_to_int,
    encode_hex,
    decode_hex as _decode_hex,
)
from ethereum.utils import (
    is_numeric,
    normalize_address as _normalize_address,
    zpad,
)


if sys.version_info.major == 2:
    integer_types = (int, long)  # NOQA
    binary_types = (bytes, bytearray)
    text_types = (unicode,)  # NOQA
    string_types = (basestring, bytearray)  # NOQA
else:
    integer_types = (int,)  # NOQA
    binary_types = (bytes, bytearray)
    text_types = (str,)
    string_types = (bytes, str, bytearray)


def is_binary(value):
    return isinstance(value, binary_types)


def is_text(value):
    return isinstance(value, text_types)


def is_string(value):
    return isinstance(value, string_types)


def is_integer(value):
    return isinstance(value, integer_types)


def is_array(value):
    return isinstance(value, (list, tuple))


def force_bytes(value):
    if is_binary(value):
        return bytes(value)
    elif is_text(value):
        return codecs.encode(value, "iso-8859-1")
    else:
        raise TypeError("Unsupported type: {0}".format(type(value)))


def force_text(value):
    if is_text(value):
        return value
    elif is_binary(value):
        return codecs.decode(value, "iso-8859-1")
    else:
        raise TypeError("Unsupported type: {0}".format(type(value)))


def force_obj_to_bytes(obj, skip_unsupported=False):
    if is_string(obj):
        return force_bytes(obj)
    elif isinstance(obj, dict):
        return {
            k: force_obj_to_bytes(v, skip_unsupported) for k, v in obj.items()
        }
    elif isinstance(obj, (list, tuple)):
        return type(obj)(force_obj_to_bytes(v, skip_unsupported) for v in obj)
    elif not skip_unsupported:
        raise ValueError("Unsupported type: {0}".format(type(obj)))
    else:
        return obj


def force_obj_to_text(obj, skip_unsupported=False):
    if is_string(obj):
        return force_text(obj)
    elif isinstance(obj, dict):
        return {
            k: force_obj_to_text(v, skip_unsupported) for k, v in obj.items()
        }
    elif isinstance(obj, (list, tuple)):
        return type(obj)(force_obj_to_text(v, skip_unsupported) for v in obj)
    elif not skip_unsupported:
        raise ValueError("Unsupported type: {0}".format(type(obj)))
    else:
        return obj


def coerce_args_to_bytes(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        bytes_args = force_obj_to_bytes(args, True)
        bytes_kwargs = force_obj_to_bytes(kwargs, True)
        return fn(*bytes_args, **bytes_kwargs)
    return inner


def coerce_return_to_bytes(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        return force_obj_to_bytes(fn(*args, **kwargs), True)
    return inner


@coerce_args_to_bytes
def strip_0x(value):
    if value.startswith(b'0x'):
        return value[2:]
    return value


@coerce_args_to_bytes
def add_0x(value):
    return b"0x" + strip_0x(value)


@coerce_args_to_bytes
def normalize_address(value, allow_blank=True):
    return _normalize_address(value, allow_blank)


@coerce_args_to_bytes
def normalize_number(value):
    if is_numeric(value):
        return value
    elif is_string(value):
        if value.startswith(b'0x'):
            return int(value, 16)
        else:
            return big_endian_to_int(value)
    else:
        raise ValueError("Unknown numeric encoding: {0}".format(value))


@coerce_return_to_bytes
def encode_address(address):
    return add_0x(encode_hex(normalize_address(address, allow_blank=True)))


@coerce_return_to_bytes
def encode_data(data, length=None):
    """Encode unformatted binary `data`.

    If `length` is given, the result will be padded like this: ``quantity_encoder(255, 3) ==
    '0x0000ff'``.
    """
    return add_0x(encode_hex(zpad(data, length or 0)))


@coerce_return_to_bytes
def encode_32bytes(value):
    return encode_data(value, 32)


@coerce_return_to_bytes
def encode_number(value, length=None):
    """Encode interger quantity `data`."""
    if not is_numeric(value):
        raise ValueError("Unsupported type: {0}".format(type(value)))
    hex_value = encode_data(int_to_big_endian(value), length)

    if length:
        return hex_value
    else:
        return add_0x(strip_0x(hex_value).lstrip(b'0') or b'0')


@coerce_args_to_bytes
def decode_hex(value):
    return _decode_hex(strip_0x(value))


def mk_random_privkey():
    return decode_hex(encode_number(random.getrandbits(256), 32))


def normalize_block_identifier(block_identifier):
    if block_identifier is None:
        return None
    elif block_identifier in ["latest", "earliest", "pending"]:
        return block_identifier
    else:
        return normalize_number(block_identifier)
